"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const command_1 = require("@oclif/command");
const fs_1 = require("fs");
const index_1 = require("../index");
const index_2 = require("../index");
const length_1 = __importDefault(require("@turf/length"));
const chalk = require('chalk');
class Intersects extends command_1.Command {
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            const { args, flags } = this.parse(Intersects);
            if (flags.out)
                this.log(chalk.bold.keyword('green')('  🌏  Loading polygon...'));
            var content = fs_1.readFileSync(args.file);
            var polygon = JSON.parse(content.toLocaleString());
            if (flags.out)
                this.log(chalk.bold.keyword('green')('  🗄️  Loading SharedStreets tiles...'));
            var params = new index_1.TilePathParams();
            params.source = 'osm/planet-180430';
            params.tileHierarchy = 6;
            var tileIndex = new index_2.TileIndex();
            if (flags.out)
                this.log(chalk.bold.keyword('green')('  🔍  Searching data...'));
            var data = yield tileIndex.intersects(polygon, index_1.TileType.GEOMETRY, 0, params);
            if (flags.stats) {
                var totalLength = 0;
                for (var feature of data.features) {
                    var segmentLength = length_1.default(feature, { units: "meters" });
                    totalLength += segmentLength;
                }
                totalLength = Math.round(totalLength * 100) / 100;
                this.log(chalk.keyword('blue')('Total Segment Length: ') + chalk.bold.keyword('blue')(totalLength));
            }
            else {
                var jsonOut = JSON.stringify(data);
                if (flags.out) {
                    fs_1.writeFileSync(flags.out, jsonOut);
                }
                else
                    this.log(jsonOut);
            }
        });
    }
}
Intersects.description = 'queries streets by polygon data and returns GeoJSON output of all intersecting features';
Intersects.examples = [
    `$ shst intersects polygon.geojson --out=output.geojson
  🌏 Loading polygon...
  🗄️ Loading SharedStreets tiles...
  🔍 Searching data...
    `,
];
Intersects.flags = {
    help: command_1.flags.help({ char: 'h' }),
    // flag with a value (-o, --out=FILE)
    out: command_1.flags.string({ char: 'o', description: 'output file' }),
    stats: command_1.flags.boolean({ char: 's' })
    // flag with no value (-f, --force)
    //force: flags.boolean({char: 'f'}),
};
Intersects.args = [{ name: 'file' }];
exports.default = Intersects;
