"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const command_1 = require("@oclif/command");
const fs_1 = require("fs");
const index_1 = require("../index");
const index_2 = require("../index");
const turfHelpers = __importStar(require("@turf/helpers"));
const geom_1 = require("../geom");
const graph_1 = require("../graph");
const envelope_1 = __importDefault(require("@turf/envelope"));
const chalk = require('chalk');
const cliProgress = require('cli-progress');
function mapOgProperties(og_props, new_props) {
    for (var prop of Object.keys(og_props)) {
        new_props['pp_' + prop] = og_props[prop];
        //console.log(new_props)
    }
}
class Match extends command_1.Command {
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            const { args, flags } = this.parse(Match);
            this.log(chalk.bold.keyword('green')('  🌏  Loading geojson data...'));
            var inFile = args.file;
            var outFile = flags.out;
            if (!inFile || !fs_1.existsSync(inFile)) {
                this.log(chalk.bold.keyword('orange')('  💾  Input file not found...'));
                return;
            }
            if (!outFile)
                outFile = inFile;
            if (outFile.toLocaleLowerCase().endsWith(".geojson") || outFile.toLocaleLowerCase().endsWith(".geojson"))
                outFile = outFile.split(".").slice(0, -1).join(".");
            if (flags['direction-field'])
                console.log(chalk.bold.keyword('green')('  Filtering one-way and two-way streets using field "' + flags['direction-field'] + '" with values: ' + ' "' + flags['one-way-with-direction-value'] + '", "' + flags['one-way-against-direction-value'] + '", "' + flags['two-way-value'] + '"'));
            var content = fs_1.readFileSync(inFile);
            var data = JSON.parse(content.toLocaleString());
            var params = new index_1.TilePathParams();
            params.source = flags['tile-source'];
            params.tileHierarchy = flags['tile-hierarchy'];
            if (data.features[0].geometry.type === 'LineString' || data.features[0].geometry.type === 'MultiLineString') {
                yield matchLines(outFile, params, data, flags);
            }
            else if (data.features[0].geometry.type === 'Point') {
                yield matchPoints(outFile, params, data, flags);
            }
        });
    }
}
Match.description = 'matches point and line features to sharedstreets refs';
Match.examples = [
    `$ shst match points.geojson --out=matched_points.geojson --port-properties
  🌏  Loading points...
  ✨  Matching 3 points...
  🎉  Matched 2 points... (1 unmached)
    `,
];
Match.flags = {
    help: command_1.flags.help({ char: 'h' }),
    // flag with a value (-o, --out=FILE)
    out: command_1.flags.string({ char: 'o', description: 'file output name creates files [file-output-name].matched.geojson and [file-output-name].unmatched.geojson' }),
    'tile-source': command_1.flags.string({ description: 'SharedStreets tile source', default: 'osm/planet-181224' }),
    'tile-hierarchy': command_1.flags.integer({ description: 'SharedStreets tile hierarchy', default: 6 }),
    'skip-port-properties': command_1.flags.boolean({ char: 'p', description: 'skip porting existing feature properties preceeded by "pp_"', default: false }),
    'follow-line-direction': command_1.flags.boolean({ description: 'only match using line direction', default: false }),
    'direction-field': command_1.flags.string({ description: 'name of optional line properity describing segment directionality, use the related "one-way-*-value" and "two-way-value" properities' }),
    'one-way-with-direction-value': command_1.flags.string({ description: 'name of optional value of "direction-field" indicating a one-way street with line direction' }),
    'one-way-against-direction-value': command_1.flags.string({ description: 'name of optional value of "direction-field" indicating a one-way street against line direction' }),
    'two-way-value': command_1.flags.string({ description: 'name of optional value of "direction-field" indicating a two-way street' }),
    'bearing-field': command_1.flags.string({ description: 'name of optional point property containing bearing in decimal degrees', default: 'bearing' }),
    'search-radius': command_1.flags.integer({ description: 'search radius for for snapping points, lines and traces', default: 10 }),
    'snap-intersections': command_1.flags.boolean({ description: 'snap line end-points to nearest intersection', default: false }),
    stats: command_1.flags.boolean({ char: 's' })
    // flag with no value (-f, --force)
    //force: flags.boolean({char: 'f'}),
};
Match.args = [{ name: 'file' }];
exports.default = Match;
function matchPoints(outFile, params, points, flags) {
    return __awaiter(this, void 0, void 0, function* () {
        console.log(chalk.bold.keyword('green')('  ✨  Matching ' + points.features.length + ' points...'));
        // test matcher point candidates
        var matcher = new index_2.PointMatcher(null, params);
        matcher.searchRadius = flags['search-radius'];
        var matchedPoints = [];
        var unmatchedPoints = [];
        for (var searchPoint of points.features) {
            var bearing = null;
            if (searchPoint.properties && searchPoint.properties[flags['bearing-field']])
                bearing = parseFloat(searchPoint.properties[flags['bearing-field']]);
            var matches = yield matcher.getPointCandidates(searchPoint, bearing, 3);
            if (matches.length > 0) {
                var matchedFeature = matches[0].toFeature();
                if (!flags['skip-port-properties'])
                    mapOgProperties(searchPoint.properties, matchedFeature.properties);
                matchedPoints.push(matchedFeature);
            }
            else {
                unmatchedPoints.push(searchPoint);
            }
        }
        if (matchedPoints.length) {
            console.log(chalk.bold.keyword('blue')('  ✏️  Writing ' + matchedPoints.length + ' matched points: ' + outFile + ".matched.geojson"));
            var matchedFeatureCollection = turfHelpers.featureCollection(matchedPoints);
            var matchedJsonOut = JSON.stringify(matchedFeatureCollection);
            fs_1.writeFileSync(outFile + ".matched.geojson", matchedJsonOut);
        }
        if (unmatchedPoints.length) {
            console.log(chalk.bold.keyword('blue')('  ✏️  Writing ' + unmatchedPoints.length + ' unmatched points: ' + outFile + ".unmatched.geojson"));
            var unmatchedFeatureCollection = turfHelpers.featureCollection(unmatchedPoints);
            var unmatchedJsonOut = JSON.stringify(unmatchedFeatureCollection);
            fs_1.writeFileSync(outFile + ".unmatched.geojson", unmatchedJsonOut);
        }
    });
}
var MatchDirection;
(function (MatchDirection) {
    MatchDirection[MatchDirection["FORWARD"] = 0] = "FORWARD";
    MatchDirection[MatchDirection["BACKWARD"] = 1] = "BACKWARD";
    MatchDirection[MatchDirection["BOTH"] = 2] = "BOTH";
})(MatchDirection || (MatchDirection = {}));
function matchLines(outFile, params, lines, flags) {
    return __awaiter(this, void 0, void 0, function* () {
        var cleanedlines = new geom_1.CleanedLines(lines);
        console.log(chalk.bold.keyword('green')('  ✨  Matching ' + cleanedlines.clean.length + ' lines...'));
        var extent = envelope_1.default(lines);
        var matcher = new graph_1.Graph(extent, params);
        yield matcher.buildGraph();
        if (flags['search-radius'])
            matcher.searchRadius = flags['search-radius'];
        matcher.snapIntersections = flags['snap-intersections'];
        var matchedLines = [];
        var unmatchedLines = [];
        const bar1 = new cliProgress.Bar({}, {
            format: chalk.keyword('blue')(' {bar}') + ' {percentage}% | {value}/{total} ',
            barCompleteChar: '\u2588',
            barIncompleteChar: '\u2591'
        });
        bar1.start(cleanedlines.clean.length, 0);
        for (var line of cleanedlines.clean) {
            const applyProperties = (path, originalFeature) => {
                path.matchedPath.properties['segments'] = path.segments;
                path.matchedPath.properties['score'] = path.score;
                path.matchedPath.properties['matchType'] = path.matchType;
                if (!flags['skip-port-properties'])
                    mapOgProperties(originalFeature.properties, path.matchedPath.properties);
                return path.matchedPath;
            };
            var matchDirection;
            if (flags['direction-field'] && line.properties[flags['direction-field'].toLocaleLowerCase()]) {
                var lineDirectionValue = '' + line.properties[flags['direction-field'].toLocaleLowerCase()];
                if (lineDirectionValue == '' + flags['one-way-with-direction-value']) {
                    matchDirection = MatchDirection.FORWARD;
                }
                else if (lineDirectionValue == '' + flags['one-way-against-direction-value']) {
                    matchDirection = MatchDirection.BACKWARD;
                }
                else if (lineDirectionValue == '' + flags['two-way-value']) {
                    matchDirection = MatchDirection.BOTH;
                }
                else {
                    // TODO handle lines that don't match rules
                    matchDirection = MatchDirection.BOTH;
                }
            }
            else if (flags['follow-line-direction']) {
                matchDirection = MatchDirection.FORWARD;
            }
            else {
                matchDirection = MatchDirection.BOTH;
            }
            var matchedOneDirection = false;
            if (matchDirection == MatchDirection.FORWARD || matchDirection == MatchDirection.BOTH) {
                var matchForward = yield matcher.match(line);
                if (matchForward) {
                    var matchedLine = applyProperties(matchForward, line);
                    matchedLines.push(matchedLine);
                    matchedOneDirection = true;
                }
            }
            if (matchDirection == MatchDirection.BACKWARD || matchDirection == MatchDirection.BOTH) {
                var reversedLine = geom_1.reverseLineString(line);
                var matchBackward = yield matcher.match(reversedLine);
                if (matchBackward) {
                    var matchedLine = applyProperties(matchBackward, reversedLine);
                    matchedLines.push(matchedLine);
                    matchedOneDirection = true;
                }
            }
            if (!matchedOneDirection)
                unmatchedLines.push(line);
            bar1.increment();
        }
        bar1.stop();
        if (matchedLines && matchedLines.length) {
            console.log(chalk.bold.keyword('blue')('  ✏️  Writing ' + matchedLines.length + ' matched edges: ' + outFile + ".matched.geojson"));
            var matchedFeatureCollection = turfHelpers.featureCollection(matchedLines);
            var matchedJsonOut = JSON.stringify(matchedFeatureCollection);
            fs_1.writeFileSync(outFile + ".matched.geojson", matchedJsonOut);
        }
        if (unmatchedLines && unmatchedLines.length) {
            console.log(chalk.bold.keyword('blue')('  ✏️  Writing ' + unmatchedLines.length + ' unmatched lines: ' + outFile + ".unmatched.geojson"));
            var unmatchedFeatureCollection = turfHelpers.featureCollection(unmatchedLines);
            var unmatchedJsonOut = JSON.stringify(unmatchedFeatureCollection);
            fs_1.writeFileSync(outFile + ".unmatched.geojson", unmatchedJsonOut);
        }
        if (cleanedlines.invalid && cleanedlines.invalid.length) {
            console.log(chalk.bold.keyword('blue')('  ✏️  Writing ' + cleanedlines.invalid + ' in lines: ' + outFile + ".invalid.geojson"));
            var invalidFeatureCollection = turfHelpers.featureCollection(cleanedlines.invalid);
            var invalidJsonOut = JSON.stringify(invalidFeatureCollection);
            fs_1.writeFileSync(outFile + ".unmatched.geojson", invalidJsonOut);
        }
    });
}
