"use strict";
//import redis = require('redis');
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
//import { SharedStreetsMetadata, SharedStreetsIntersection, SharedStreetsGeometry, SharedStreetsReference, RoadClass  } from 'sharedstreets-types';
const turfHelpers = __importStar(require("@turf/helpers"));
const buffer_1 = __importDefault(require("@turf/buffer"));
const along_1 = __importDefault(require("@turf/along"));
const line_slice_along_1 = __importDefault(require("@turf/line-slice-along"));
const geojson_rbush_1 = __importDefault(require("geojson-rbush"));
const index_1 = require("../src/index");
const tiles_1 = require("./tiles");
const SHST_ID_API_URL = 'https://api.sharedstreets.io/v0.1.0/id/';
// maintains unified spaital and id indexes for tiled data
function createIntersectionGeometry(data) {
    var point = turfHelpers.point([data.lon, data.lat]);
    return turfHelpers.feature(point.geometry, { id: data.id });
}
function createGeometry(data) {
    var line = turfHelpers.lineString(index_1.lonlatsToCoords(data.lonlats));
    var feature = turfHelpers.feature(line.geometry, { id: data.id });
    return feature;
}
class TileIndex {
    constructor() {
        this.tiles = new Set();
        this.objectIndex = new Map();
        this.featureIndex = new Map();
        this.intersectionIndex = geojson_rbush_1.default(9);
        this.geometryIndex = geojson_rbush_1.default(9);
    }
    isIndexed(tilePath) {
        if (this.tiles.has(tilePath.toPathString()))
            return true;
        else
            return false;
    }
    indexTilesByPathGroup(tilePathGroup) {
        return __awaiter(this, void 0, void 0, function* () {
            for (var tilePath of tilePathGroup) {
                yield this.indexTileByPath(tilePath);
            }
            return false;
        });
    }
    indexTileByPath(tilePath) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isIndexed(tilePath))
                return true;
            var data = yield tiles_1.getTile(tilePath);
            if (tilePath.tileType === tiles_1.TileType.GEOMETRY) {
                var geometryFeatures = [];
                for (var geometry of data) {
                    if (!this.objectIndex.has(geometry.id)) {
                        this.objectIndex.set(geometry.id, geometry);
                        var geometryFeature = createGeometry(geometry);
                        this.featureIndex.set(geometry.id, geometryFeature);
                        geometryFeatures.push(geometryFeature);
                    }
                }
                this.geometryIndex.load(geometryFeatures);
            }
            else if (tilePath.tileType === tiles_1.TileType.INTERSECTION) {
                var intersectionFeatures = [];
                for (var intersection of data) {
                    if (!this.objectIndex.has(intersection.id)) {
                        this.objectIndex.set(intersection.id, intersection);
                        var intesectionFeature = createIntersectionGeometry(intersection);
                        this.featureIndex.set(intersection.id, intesectionFeature);
                        intersectionFeatures.push(intesectionFeature);
                    }
                }
                this.intersectionIndex.load(intersectionFeatures);
            }
            else if (tilePath.tileType === tiles_1.TileType.REFERENCE) {
                for (var reference of data) {
                    this.objectIndex.set(reference.id, reference);
                }
            }
            else if (tilePath.tileType === tiles_1.TileType.METADATA) {
                for (var metadata of data) {
                    this.metadataIndex.set(metadata.geometryId, metadata.osmMetadata);
                }
            }
            this.tiles.add(tilePath.toPathString());
        });
    }
    getGraph(polygon, params) {
        return __awaiter(this, void 0, void 0, function* () {
            return null;
        });
    }
    intersects(polygon, searchType, buffer, params, additionalTypes = null) {
        return __awaiter(this, void 0, void 0, function* () {
            var tilePaths = tiles_1.TilePathGroup.fromPolygon(polygon, buffer, params);
            if (searchType === tiles_1.TileType.GEOMETRY)
                tilePaths.addType(tiles_1.TileType.GEOMETRY);
            else if (searchType === tiles_1.TileType.INTERSECTION)
                tilePaths.addType(tiles_1.TileType.INTERSECTION);
            else
                throw "invalid search type must be GEOMETRY or INTERSECTION";
            if (additionalTypes) {
                for (var type of additionalTypes) {
                    tilePaths.addType(type);
                }
            }
            yield this.indexTilesByPathGroup(tilePaths);
            var data;
            if (searchType === tiles_1.TileType.GEOMETRY)
                data = this.geometryIndex.search(polygon);
            else if (searchType === tiles_1.TileType.INTERSECTION)
                data = this.intersectionIndex.search(polygon);
            return data;
        });
    }
    nearby(point, searchType, searchRadius, params, additionalTypes = null) {
        return __awaiter(this, void 0, void 0, function* () {
            var tilePaths = tiles_1.TilePathGroup.fromPoint(point, searchRadius * 2, params);
            if (searchType === tiles_1.TileType.GEOMETRY)
                tilePaths.addType(tiles_1.TileType.GEOMETRY);
            else if (searchType === tiles_1.TileType.INTERSECTION)
                tilePaths.addType(tiles_1.TileType.INTERSECTION);
            else
                throw "invalid search type must be GEOMETRY or INTERSECTION";
            if (additionalTypes) {
                for (var type of additionalTypes) {
                    tilePaths.addType(type);
                }
            }
            yield this.indexTilesByPathGroup(tilePaths);
            var bufferedPoint = buffer_1.default(point, searchRadius, { 'units': 'meters' });
            var data;
            if (searchType === tiles_1.TileType.GEOMETRY)
                data = this.geometryIndex.search(bufferedPoint);
            else if (searchType === tiles_1.TileType.INTERSECTION)
                data = this.intersectionIndex.search(bufferedPoint);
            return data;
        });
    }
    geom(referenceId, p1, p2) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.objectIndex.has(referenceId)) {
                var ref = this.objectIndex.get(referenceId);
                var geom = this.objectIndex.get(ref.geometryId);
                var geomFeature = JSON.parse(JSON.stringify(this.featureIndex.get(ref.geometryId)));
                if (geom.backReferenceId && geom.backReferenceId === referenceId) {
                    geomFeature.geometry.coordinates = geomFeature.geometry.coordinates.reverse();
                }
                if (p1 < 0)
                    p1 = 0;
                if (p2 < 0)
                    p2 = 0;
                if (p1 == null && p2 == null) {
                    return geomFeature;
                }
                else if (p1 && p2 == null) {
                    return along_1.default(geomFeature, p1, { "units": "meters" });
                }
                else if (p1 != null && p2 != null) {
                    try {
                        return line_slice_along_1.default(geomFeature, p1, p2, { "units": "meters" });
                    }
                    catch (e) {
                        //console.log(p1, p2)
                    }
                }
            }
            // TODO find missing IDs via look up
            return null;
        });
    }
}
exports.TileIndex = TileIndex;
